package com.loafle.overflow.container.general.crawler.impl.jmx;

import com.loafle.overflow.crawler.Crawler;
import com.loafle.overflow.model.sensorconfig.Item;
import com.loafle.overflow.model.sensorconfig.ResultSet;
import com.loafle.overflow.model.sensorconfig.SensorConfig;
import com.loafle.overflow.core.exception.OverflowException;

import javax.management.*;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;
import javax.management.remote.rmi.RMIConnectorServer;
import javax.naming.Context;
import javax.rmi.ssl.SslRMIClientSocketFactory;
import java.io.IOException;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

public class JMXCrawler implements Crawler {
  private static final Logger logger = Logger.getLogger(JMXCrawler.class.getName());

  public JMXCrawler() {
  }

  @Override
  public String name() {
    return "JMX";
  }

  @Override
  public String toString() {
    return "JMX Crawler";
  }

  @Override
  public void auth(Map<String, Object> authInfoMap) throws OverflowException {
    String hostIp = (String) authInfoMap.get("ip");
    Integer port = (Integer) authInfoMap.get("port");

    String jmxUrl = "service:jmx:rmi:///jndi/rmi://" + hostIp + ":" + port + "/jmxrmi";
    String username = (String) authInfoMap.get("id");
    String password = (String) authInfoMap.get("pw");
    Boolean useSSL = (Boolean) authInfoMap.get("ssl");

    // connection
    JMXConnector jmxconn = null;
    try {
      Map<String, Object> environment = getAuthority(useSSL, username, password);
      jmxconn = JMXConnectorFactory.connect(new JMXServiceURL(jmxUrl), environment);
      jmxconn.getMBeanServerConnection();
    } catch (Exception e) {
      throw new OverflowException("JMXCrawler", e);
    } finally {
      if (jmxconn != null) {
        try {
          jmxconn.close();
        } catch (IOException e) {
        }
      }
    }
  }

  @Override
  public Map<String, String> get(SensorConfig sensorConfig) throws OverflowException {
    String hostIp = sensorConfig.getTarget().getConnection().getIp();
    int port = sensorConfig.getTarget().getConnection().getPort();

    String jmxUrl = "service:jmx:rmi:///jndi/rmi://" + hostIp + ":" + port + "/jmxrmi";
    String username = (String) sensorConfig.getTarget().getAuth().get("id");
    String password = (String) sensorConfig.getTarget().getAuth().get("pw");
    boolean useSSL = sensorConfig.getTarget().getConnection().isSsl();

    // connection
    MBeanServerConnection beanCon;
    JMXConnector jmxconn = null;
    try {
      Map<String, Object> environment = getAuthority(useSSL, username, password);
      jmxconn = JMXConnectorFactory.connect(new JMXServiceURL(jmxUrl), environment);
      beanCon = jmxconn.getMBeanServerConnection();

      Map<String, String> resultMap = new HashMap<>();

      List<ObjectName> objectNames = new LinkedList<ObjectName>();
      objectNames.add(null);
      ObjectName on = null;

      for (Item item : sensorConfig.getItems()) {

        ResultSet resultSet = ResultSet.newInstance(item);

        // get object
        String query = item.getQueryInfo().getQuery();
        on = new ObjectName(query);
        Set<ObjectInstance> names = beanCon.queryMBeans(on, null);

        List<Map<String, Object>> keys = (List<Map<String, Object>>) item.getQueryInfo().getExtend().get("aliases");

        for (ObjectInstance name : names) {

          Map<String, String> attrs = parseAttribute(beanCon, name.getObjectName());
          if (keys != null) {
            for (Map<String, Object> keyMap : keys) {
              String key = (String) keyMap.get("key");
              int index = (int) keyMap.get("index");
              String objectName = parseObjectName(name.toString(), key);
              String alias = item.getMappingInfo().getArrayColumns().get(index);
              attrs.put(alias, objectName);
            }
          }

          resultSet.addRow(parse(attrs, resultSet.getMeta()));
        }

        resultMap.putAll(resultSet.getData());
      }

      return resultMap;
    } catch (Exception e) {
      throw new OverflowException("JMXCrawler", e);
    } finally {
      if (jmxconn != null) {
        try {
          jmxconn.close();
        } catch (IOException e) {
        }
      }
    }
  }

  private List<String> parse(Map<String, String> dataMap, Map<String, Integer> meta) {

    List<String> row = new ArrayList<>(Arrays.asList(new String[meta.size()]));
    for (Map.Entry<String, Integer> info : meta.entrySet()) {
      String data = dataMap.get(info.getKey());
      data = data.trim();
      row.set(info.getValue().intValue(), data);
    }
    return row;
  }

  private Map<String, String> parseAttribute(MBeanServerConnection beanCon, ObjectName mbeanName)
      throws IntrospectionException, ReflectionException, InstanceNotFoundException, IOException,
      AttributeNotFoundException, MBeanException {

    Map<String, String> returnMap = new HashMap<>();
    MBeanInfo info = beanCon.getMBeanInfo(mbeanName);

    for (MBeanAttributeInfo attr : info.getAttributes()) {

      if (!attr.isReadable()) {
        logCrawler(mbeanName, attr, "not readable");
        continue;
      }

      //            System.out.println("========================================================================");
      //            System.out.println("PropertyListString"+mbeanName.getCanonicalKeyPropertyListString());
      //            System.out.println("canonicaName="+mbeanName.getCanonicalName());
      //            System.out.println(mbeanName.getDomain());
      //            System.out.println(mbeanName.getKeyPropertyListString().compareTo("http"));
      //            System.out.println("========================================================================");
      Object att = null;

      try {
        att = beanCon.getAttribute(mbeanName, attr.getName());
      } catch (Exception e) {
        logCrawler(mbeanName, attr, "fail: " + e);
        continue;
      }

      if (att != null)
        returnMap.put(attr.getName(), att.toString());
    }
    return returnMap;
  }

  private String parseObjectName(String s, String key) {

    int start = s.indexOf("[");
    int end = s.lastIndexOf("]");
    s = s.substring(start + 1, end);

    String[] slices = s.split(",");
    String returnData = "";
    for (String item : slices) {
      String[] itemSplit = item.split("=");
      if (itemSplit == null || itemSplit.length != 2) {
        continue;
      }
      if (itemSplit[0].equals(key)) {
        returnData = itemSplit[1];
      }
    }

    return returnData;
  }

  private Map<String, Object> getAuthority(boolean useSSL, String username, String password) {
    Map<String, Object> environment = new HashMap<String, Object>();

    if (username != null && username.length() != 0 && password != null && password.length() != 0) {
      String[] credent = new String[] { username, password };
      environment.put(JMXConnector.CREDENTIALS, credent);
    }

    if (useSSL) {
      environment.put(Context.SECURITY_PROTOCOL, "ssl");
      SslRMIClientSocketFactory clientSocketFactory = new SslRMIClientSocketFactory();
      environment.put(RMIConnectorServer.RMI_CLIENT_SOCKET_FACTORY_ATTRIBUTE, clientSocketFactory);
      environment.put("com.sun.jndi.rmi.factory.socket", clientSocketFactory);
    }

    return environment;
  }

  private static void logCrawler(String name, String msg) {
    logger.log(Level.FINE, "crawler : '" + name + "': " + msg);
  }

  private static void logCrawler(ObjectName mbeanName, MBeanAttributeInfo attr, String msg) {
    logCrawler(mbeanName + "'_'" + attr.getName(), msg);
  }

}
